//: Playground - noun: a place where people can play

import UIKit

var str = "Hello, playground"

// MARK: - Node

public class Node<T> {
    typealias NodeType = Node<T>                            // just to shorten definitions
    var value: T                                            // that's the object we want to keep in the list
    var next: NodeType?                                     // since this is a class, that's the pointer to the next node
    weak var previous: NodeType?                            // and thats the one the previous node
    var nextOrFirst: NodeType? {                            // a computed variable to make endless loops possible
        if next == nil {                                    // we reached the end of the list
            if let l = list {                               // just to be safe
                return l.head                               // start again from the top
            } else {                                        // must have unlinked the object
                return nil                                  // can't do anything with it
            }
        } else {                                            // not reached end yet
            return next                                     // take the next node
        }
    }

    fileprivate weak var list: LinkedList<T>?               // let's keep a pointer to the list as well

    init(value: T, list: LinkedList<T>) {                   // let's initialize our node
        self.value = value                                  // that's the object we want to keep
        self.list = list                                    // and remember our list as well
    }

    public func removeFromList() -> T {                     // let's remove this node from the list
        if let l = list {                                   // yes, we do know our list
            return l.remove(node: self)                     // let the list do it's job
        } else {                                            // oh no, no list known (any more)
            return value                                    // just return our object
        }
    }
}

// MARK: - Iterator

public struct LinkedListIterator<T>: IteratorProtocol {     // let's create an iterator for our linked list
    public typealias Element = Node<T>                      // the type of each node in the list
    
    private var currentNode: Element?                       // the current node in the iteration
    
    fileprivate init(startNode: Element?) {                 // let's initiate the iterator
        currentNode = startNode                             // just set the current node
    }
    
    public mutating func next() -> LinkedListIterator.Element? {    // get the next element in the list
        let node = currentNode                              // take the current node
        currentNode = currentNode?.next                     // and remember the next one, so the current node may
                                                            // be invalidated in the iteration
        return node                                         // return the node (not only the object, i.e. node?.value)
    }
}

// MARK: - LinkedList

public class LinkedList<T>: Sequence {                      // the linked list, finally, conforming to the sequence protocol
    
    public typealias Iterator = LinkedListIterator<T>       // the type of our iterator
    
    public func makeIterator() -> LinkedList.Iterator {     // let's create our iterator
        return LinkedListIterator(startNode: head)          // initialise it with the head of our list
    }
    
    typealias NodeType = Node<T>                            // the type of our nodes
    fileprivate var head: NodeType?                         // the first element of our list
    private var tail: NodeType?                             // and the last one; both may be nil
    
    public fileprivate(set) var count: Int = 0              // the number of elements in the list at any given time
    
    public init() {                                         // just needed to add further inits
    }

    public init<S: Sequence>(_ elements: S) where S.Iterator.Element == T {     // create the list from an other sequence
        
        for element in elements {                           // iterate over all elements in the supplied sequence
            self.append(value: element)                     // and extend our list
        }
    }

    public var isEmpty: Bool {                              // convenience property to test for emptiness
        return head == nil                                  // if the head is nil the list is empty
    }
    
    public var first: Node<T>? {                            // get the head of the list
        return head
    }
    
    public var last: Node<T>? {                             // and its tail
        return tail
    }
    
    public func append(value: T) {                          // append a new node at the end of the list
        let newNode = Node(value: value, list: self)        // create the node with the object
        if let tailNode = tail {                            // we do have an end (list is not empty)
            newNode.previous = tailNode                     // remember that node in our new one
            tailNode.next = newNode                         // and make the new node the new tail
        } else {                                            // the list is empty
            head = newNode                                  // make our new node the head of the list
        }
        tail = newNode                                      // our new node is the (new) end of the list
        count += 1                                          // and increase our counter
    }
    
    public func nodeAt(index: Int) -> Node<T>? {            // get the node at the specified index
        if index >= 0 {                                     // make sure the index is positive
            var node = head                                 // start at the head of the list
            var i = index                                   // the count down
            while node != nil {                             // iterate over the list
                if i == 0 { return node }                   // until we finished the countdown and return the node
                i -= 1                                      // countdown
                node = node!.next                           // and take the next item
            }
        }
        return nil                                          // index is out of range
    }
    
    public subscript(index: Int) -> T {                     // allow the list to be subscripted
        let node = nodeAt(index: index)                     // get the node at the specified index
        assert(node != nil)                                 // make sure, its not nil
        return node!.value                                  // and return its value
    }
    
    public func indexOf(node n: Node<T>) -> Int? {          // get the index of a supplied node in the list
        var i = 0                                           // start at the beginning
        var node = head                                     // with the head of the list
        while node != nil {                                 // and iterate over the whole list
            if node! === n { return i }                     // until we find the identical node: return its index
            i += 1                                          // increase the index
            node = node!.next                               // and take the next item
        }
        return nil                                          // could not find the supplied node
    }

    public func removeAll() {                               // empty the list
        head = nil                                          // clear its head
        tail = nil                                          // and its tail
        count = 0                                           // and reset the counter
    }
    
    public func remove(node: Node<T>) -> T {                // remove a node from the list
        let prev = node.previous                            // first get its neighbours
        let next = node.next
        
        if let prev = prev {                                // we are not at the top of the list
            prev.next = next                                // make the previous node point to our following one
        } else {                                            // we are at the top of the list
            head = next                                     // make our follower the new head of the list
        }
        next?.previous = prev                               // and make our follower point to our previous node
        
        if next == nil {                                    // we are at the end of the list and do not have a follower
            tail = prev                                     // make the previous node the new tail of the list
        }
        
        node.previous = nil                                 // and clear the pointers in our node
        node.next = nil
        count -= 1                                          // decrease the counter
        
        return node.value                                   // and return our object
    }
    
    private func nodesBeforeAndAfter(index: Int) -> (NodeType?, NodeType?) {    // get the nodes before and after an index
        assert(index >= 0)                                  // make sure the index is positive (+ 0)
        
        var i = index                                       // let's countdown
        var next = head                                     // start at the top of the list
        var prev: NodeType?                                 // in the beginning the previous node is nil
        
        while next != nil && i > 0 {                        // iterate over the list
            i -= 1                                          // and count down
            prev = next                                     // the previous node becomes the current one
            next = next!.next                               // and the current one the next
        }
        assert(i == 0)                                      // make sure we hit the specified index
        
        return (prev, next)                                 // and return the previous and the current node
    }
    
    private func insert(value: T, prev: NodeType?, next: NodeType?) {   // insert a new node between neighbours
        let newNode = Node(value: value, list: self)        // create the new node from the object
        newNode.previous = prev                             // let's point to the old neighbours
        newNode.next = next
        prev?.next = newNode                                // and from the neighbours to our new node
        next?.previous = newNode
        
        if prev == nil {                                    // we inserted at the top of the list
            head = newNode                                  // the head will be our new node
        }
        if next == nil {                                    // we inserted at the end of the list
            tail = newNode                                  // the tail will be our new node
        }
        count += 1                                          // and increase the counter
    }
    
    public func insert(value: T, atIndex index: Int) {      // insert an object at a specified index
        let (prev, next) = nodesBeforeAndAfter(index: index)    // get the neighbours at the index
        insert(value: value, prev: prev, next: next)        // and insert it in between
    }

    public func insert(value: T, afterNode node: Node<T>?) {    // insert an object after a supplied node
        if node == nil {                                    // we want to insert at the top of the list
            insert(value: value, atIndex: 0)                // make the new node be index 0
        } else {                                            // not at the top of the list
            insert(value: value, prev: node, next: node!.next)  // insert it between the supplied node and its follower
        }
    }
    
    public func insert(value: T, beforeNode node: Node<T>?) {   // insert an object before a supplied node
        if node == nil {                                    // we want to insert at the end of the list
            append(value: value)                            // just append the object
        } else {                                            // not at the end of the list
            insert(value: value, prev: node!.previous, next: node)  // insert it between the previous node and its follower
        }
    }
}

extension LinkedList {
    public var description: String {
        var text = "["
        var node = head
        while node != nil {
            text += "\(node!.value)"
            node = node!.next
            if node != nil { text += ", " }
        }
        return text + "]"
    }
}

let stringArray = ["1. item", "2. item", "3. item"]
let mySecondList = LinkedList<String>(stringArray)
Swift.print("The second list: \(mySecondList.description)")

for item in mySecondList {
    Swift.print("Item: \(item.value)")
    if item.value == "2. item" {
        mySecondList.insert(value: "Item 1b", beforeNode: item)
        mySecondList.insert(value: "Item 2b", afterNode: item)
    }
}
Swift.print("The second list, modified: \(mySecondList.description)")

Swift.print("---")
for item in mySecondList {
    if item.value == "2. item" {
        Swift.print("Item: \(item.value) at index \(mySecondList.indexOf(node: item)!)")
    }
}

Swift.print("---")
for item in mySecondList {
    Swift.print("Item: \(item.value)")
    if item.value == "2. item" {
        item.removeFromList()
    }
}
Swift.print("The second list, further modified: \(mySecondList.description)")

Swift.print("---")
var item = mySecondList.nodeAt(index: 2)
for _ in 0 ..< mySecondList.count {
    Swift.print("\(item!.value)")
    item = item!.nextOrFirst
}
